// Copyright © 2024 ESWIN. All rights reserved.
//
// Beijing ESWIN Computing Technology Co., Ltd and its affiliated companies ("ESWIN") retain
// all intellectual property and proprietary rights in and to this software. Except as expressly
// authorized by ESWIN, no part of the software may be released, copied, distributed, reproduced,
// modified, adapted, translated, or created derivative work of, in whole or in part.

#include <iostream>
#include <fstream>
#include <cstring>
#include <filesystem>
#include "es_npu_interface.h"
#include "preprocess.h"
#include "cJSON.h"

#define GRAY_PADDING_VALUE 114

#define DEBUG_MODE 0

#if DEBUG_MODE
#define DEBUG_INFO(fmt, ...) printf("INFO: " fmt "\n", ##__VA_ARGS__)
#define DEBUG_ERROR(fmt, ...) printf("ERROR: " fmt "\n", ##__VA_ARGS__)
#else
#define DEBUG_INFO(fmt, ...)
#define DEBUG_ERROR(fmt, ...) printf("ERROR: " fmt "\n", ##__VA_ARGS__)
#endif

EsPreProcess::EsPreProcess() {}

EsPreProcess::~EsPreProcess() {}

bool EsPreProcess::initializeFromJson(const std::string &cfgFilePath)
{
    std::ifstream file(cfgFilePath);
    if (!file.is_open()) {
        DEBUG_ERROR("Failed to open JSON file\n");
        return false;
    }

    std::string jsonString((std::istreambuf_iterator<char>(file)), std::istreambuf_iterator<char>());

    cJSON *root = cJSON_Parse(jsonString.c_str());
    if (!root) {
        DEBUG_ERROR("Failed to parse JSON\n");
        return false;
    }

    cJSON *quant = cJSON_GetObjectItem(root, "quant");
    cJSON *meanItem = cJSON_GetObjectItem(quant, "mean");
    cJSON *stdItem = cJSON_GetObjectItem(quant, "std");

    if (!meanItem || !stdItem) {
        DEBUG_ERROR("Failed to get mean and std\n");
        cJSON_Delete(root);
        return false;
    }

    if (cJSON_GetArraySize(meanItem) == 3) {
        mean = cv::Scalar(cJSON_GetArrayItem(meanItem, 0)->valuedouble, cJSON_GetArrayItem(meanItem, 1)->valuedouble,
                          cJSON_GetArrayItem(meanItem, 2)->valuedouble);
    } else {
        mean = cv::Scalar(123.675, 116.28, 103.53);
    }

    if (cJSON_GetArraySize(stdItem) == 3) {
        std = cv::Scalar(cJSON_GetArrayItem(stdItem, 0)->valuedouble, cJSON_GetArrayItem(stdItem, 1)->valuedouble,
                         cJSON_GetArrayItem(stdItem, 2)->valuedouble);
    } else {
        std = cv::Scalar(0.01712475, 0.017507, 0.01742919);
    }

    cJSON *preprocess = cJSON_GetObjectItem(root, "preprocess");
    input_format = cJSON_GetObjectItem(preprocess, "input_format")->valuestring;
    keep_ratio = cJSON_GetObjectItem(preprocess, "keep_ratio")->valueint;
    channel_first = cJSON_GetObjectItem(preprocess, "channel_first")->valueint;
    norm = cJSON_GetObjectItem(preprocess, "norm")->valueint;
    resize_shape = cJSON_GetArrayItem(cJSON_GetObjectItem(preprocess, "resize_shape"), 0)->valueint;
    crop_width = cJSON_GetArrayItem(cJSON_GetObjectItem(preprocess, "crop_shape"), 0)->valueint;
    crop_height = cJSON_GetArrayItem(cJSON_GetObjectItem(preprocess, "crop_shape"), 1)->valueint;
    quantization_scale = cJSON_GetObjectItem(quant, "quantized_scale")->valuedouble;
    // quantization_scale = 0.020787401574803147;

    DEBUG_INFO("===============================================\n");
    DEBUG_INFO("Parsed Preprocess Configuration From JSON:\n");
    DEBUG_INFO("mean: [%f, %f, %f]\n", mean[0], mean[1], mean[2]);
    DEBUG_INFO("std: [%f, %f, %f]\n", std[0], std[1], std[2]);
    DEBUG_INFO("input_format: %s\n", input_format.c_str());
    DEBUG_INFO("keep_ratio: %s\n", keep_ratio ? "true" : "false");
    DEBUG_INFO("channel_first: %s\n", channel_first ? "true" : "false");
    DEBUG_INFO("norm: %s\n", norm ? "true" : "false");
    DEBUG_INFO("resize_shape: %d\n", resize_shape);
    DEBUG_INFO("crop_shape: [%d, %d]\n", crop_width, crop_height);
    DEBUG_INFO("quantization_scale: %f\n", quantization_scale);
    DEBUG_INFO("===============================================\n");

    cJSON_Delete(root);

    return true;
}

int32_t EsPreProcess::preprocess(const std::string &jpegPath, uint8_t *inputBuf, int32_t size)
{
    cv::Mat img = cv::imread(jpegPath);
    if (img.empty()) {
        DEBUG_ERROR("Cannot read the image: %s\n", jpegPath.c_str());
        return ES_FAILURE;
    }

    // Convert to RGB as mean and std is RGB
    cv::cvtColor(img, img, cv::COLOR_BGR2RGB);

    // Get original dimensions
    int originalHeight = img.rows;
    int originalWidth = img.cols;

    if (keep_ratio) {
        // Calculate scaling factor
        double scale = static_cast<double>(resize_shape) / std::max(originalHeight, originalWidth);

        // Calculate new dimensions
        int newHeight = static_cast<int>(originalHeight * scale);
        int newWidth = static_cast<int>(originalWidth * scale);

        // Resize the image
        cv::Mat imgResized;
        cv::resize(img, imgResized, cv::Size(newWidth, newHeight));

        // Create padded background
        cv::Mat paddedImg = cv::Mat::zeros(resize_shape, resize_shape, imgResized.type()) +
                            cv::Scalar(GRAY_PADDING_VALUE, GRAY_PADDING_VALUE, GRAY_PADDING_VALUE);

        // Calculate padding position
        int startX = (resize_shape - newWidth) / 2;
        int startY = (resize_shape - newHeight) / 2;

        // Paste the resized image onto the padded background
        imgResized.copyTo(paddedImg(cv::Rect(startX, startY, newWidth, newHeight)));
        img = paddedImg;
    } else {
        cv::resize(img, img, cv::Size(resize_shape, resize_shape));
    }

    // Center crop
    int cropStartX = (resize_shape - crop_width) / 2;
    int cropStartY = (resize_shape - crop_height) / 2;
    cv::Rect cropRect(cropStartX, cropStartY, crop_width, crop_height);
    cv::Mat imgCropped = img(cropRect);

    // Convert to float type and normalize
    imgCropped.convertTo(imgCropped, CV_32F, 1.0 / 255.0);
    cv::subtract(imgCropped, mean, imgCropped);
    cv::divide(imgCropped, std, imgCropped);

    /////////////////////////////////////////////////////////////////////////////////////////////

    if (input_format == "BGR") {
        cv::cvtColor(imgCropped, imgCropped, cv::COLOR_RGB2BGR);
    }

#if DEBUG_MODE
    // Write preprocess data to jpeg
    cv::Mat preprocessedImgToSave;
    imgCropped.convertTo(preprocessedImgToSave, CV_8U, 255.0);
    imwrite("preprocess.jpeg", preprocessedImgToSave);
#endif

    /////////////////////////////////////////////////////////////////////////////////////////////

    // Convert to quantized int8
    cv::Mat quantized_img;
    imgCropped.convertTo(quantized_img, CV_32S, 1.0 / quantization_scale);
    quantized_img = min(max(quantized_img, -128), 127);

    cv::Mat int8Img;
    quantized_img.convertTo(int8Img, CV_8S);
    if (channel_first) {
        // If c first, convert to NCHW format, opencv default format is NHWC
        std::vector<cv::Mat> channels(3);
        cv::split(int8Img, channels);
        cv::Mat nchwImg;
        cv::vconcat(channels, nchwImg);
        nchwImg = nchwImg.reshape(1, std::vector<int>{1, 3, crop_width, crop_height});
        int8Img = nchwImg;
    }

    int32_t outDataLen = int8Img.total() * int8Img.elemSize();
    DEBUG_INFO("preprocess buffer size:%d, inTensor bufferSize:%d\n", outDataLen, size);
    if (outDataLen > size) {
        DEBUG_ERROR("invalid out buffer size\n");
        return ES_FAILURE;
    }
#if DEBUG_MODE
    std::filesystem::path filePath(jpegPath);
    std::string fileName = filePath.filename().string();
    std::string saveFile = fileName + ".bin";
    std::ofstream outFileStream(saveFile, std::ios::out | std::ios::binary);
    outFileStream.write((char*)int8Img.data, outDataLen);
    outFileStream.close();
#endif

    memcpy(inputBuf, int8Img.data, outDataLen);

    return ES_SUCCESS;
}
